# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Tools for manipulating and fitting kinetic functions"""
# python 3 compatibility
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import numpy as np
from spyks import core


def ode_fun(model, var, fname="inf"):
    """Create a lambda to evaluate var_\inf(V) or var_\tau(V) or var_\alpha(V) or var_\beta(V)"""
    from sympy import Symbol, lambdify
    ode = core.kinetic_ode(var, **model["eqns_unparsed"][var])
    context = {Symbol(n): v.magnitude for n, v in model["parameters"]}
    return lambdify([Symbol('V')], getattr(ode, fname).subs(context))


def tanh_inf(p, V):
    vm, dvm = p[:2]
    return (1 + np.tanh((V - vm) / dvm)) / 2


def tanh_tau(p, V):
    tm0, tm1, vmt, dvmt = p[:4]
    return (tm0 + tm1 * (1 - np.tanh((V - vmt)/dvmt)**2))


def tanh_tau2(p, V):
    tm0, tm1, vmt, dvmt1, tm2, dvmt2 = p[:6]
    return (tm0 + (1 + tm1 * np.tanh((V - vmt)/dvmt1) * (1 - tm2 * np.tanh((V - vmt)/dvmt2))))


def tanh_tau3(p, V):
    tm0, tm1, vmt, dvmt, deltt = p[:6]
    return (tm0 + deltt + 0.5 * (1 - np.tanh(V - vmt)) * tm1 * (1 - np.tanh((V - vmt) / dvmt)**2 - deltt))


def exp_inf(p, V):
    nam_v, nam_dv, pow = p[:3]
    return np.power((1 + np.exp(-(V - nam_v) / nam_dv)), pow)


def exp_tau(p, V):
    nam_t0, nam_t1, nam_tv, nam_tdv1, nam_t2, nam_tdv2 = p[:6]
    return (nam_t0 + 1/(nam_t1 * np.exp((V - nam_tv) / nam_tdv1) +
                        nam_t2 * np.exp(-(V - nam_tv) / nam_tdv2)))


def refit_kinetics(V, modelfun, candidatefun, p0):
    """Refit the kinetic function modelfun to candidatefun.

    For example, if the kinetics are specified using some funky \alpha(V) and
    \beta(V) formulation, you can estimate the best parameters for an x_\inf(V)
    and \tau(V) formulation based on hyperbolic tangents.

    V - the voltage values at which to evaluate the function
    modelfun - a callable f(V) that evaluates the target function over V
    candidatefun - a callable f(p, V) that evaluates the candidate function over V with params p
    p0 - initial parameter guess
    """
    from scipy import optimize
    errf = lambda p, x, y: candidatefun(p, x) - y
    return optimize.leastsq(errf, p0[:], args=(V, modelfun(V)))



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""
Functions to validate models, primarily with respect to dimensional correctness.
"""
# python 3 compatibility
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals


def context(model):
    """ Returns the model's components and related math functions in a dict for eval'ing equations """
    import math
    import numpy as nx
    context = {'math': math, 'exp': nx.exp, 'tanh': nx.tanh, 'pow': nx.power, 'sqrt': nx.sqrt, 'expm1': nx.expm1}
    context.update(model['state'])
    context.update(model['forcing'])
    context.update(model['parameters'])
    return context


def evalf(eqn, context):
    from sympy.printing.lambdarepr import lambdarepr
    return eval(lambdarepr(eqn), {}, context)


def _deriv_g(model):
    """ Generator that evaluates each component of the model system dX/dt = F(X, t). Will throw error if dimensional analysis fails. """


def deriv(model):
    """ Evaluates the model system dX/dt = F(X, t). Will throw error if dimensional analysis fails. """
    ctx = context(model)
    return [(variable, evalf(eqn, ctx)) for variable, eqn in model['equations']]


def reset(model):
    """ Returns the model's post-reset state, or False if it did not reset """
    ctx = context(model)
    if evalf(model['reset']['predicate'], ctx):
        return [(variable, evalf(eqn, ctx)) for variable, eqn in model['reset']['state']]
    else:
        return False


def check_symbols(model):
    """ Raises ValueError iff mismatch between equations and parameter/state/forcing vectors"""
    from itertools import chain
    from .core import symbols
    lhs, rhs = symbols(model)
    vars = {n for n, v in chain(
        model["state"], model["parameters"], model["forcing"])}
    d = vars.symmetric_difference(str(s) for s in rhs)
    if len(d) > 0:
        raise ValueError("unmatched variables in model: {}".format(d))


def check_equations(model):
    """ Raises ValueError iff mismatch between names or units of equations and state vector """
    from pint.errors import DimensionalityError
    from .core import ureg
    ctx = context(model)
    dt = 1 * ureg.ms
    X = model['state']
    dX = list()
    # check dimensions
    for variable, eqn in model['equations']:
        try:
            d = evalf(eqn, ctx)
            dX.append((variable, d))
        except DimensionalityError as e:
            raise ValueError(
                "Dimensionality error in d{}/dt = {}: {}".format(variable, eqn, e))
    for ((n1, v1), (n2, v2)) in zip(X, dX):
        if n1 != n2:
            raise ValueError(
                "Name mismatch between state ({}) and derivative ({})".format(n1, n2))
        try:
            x = v1 + dt * v2
        except:
            raise ValueError("Units mismatch between ({}) and derivative ({}) for {}".format(v1.units,
                                                                                             v2.units,
                                                                                             n1))


def check_forcing(model, forcing):
    """ Raises ValueError iff mismatch between number of components in forcing array and model """
    if forcing.ndim > 2:
        raise ValueError("Forcing must be 1-D or 2-D")
    N = len(model['forcing'])
    D = forcing.shape[1] if forcing.ndim > 1 else 1
    if N != D:
        raise ValueError(
            "Forcing component count ({}) does not match model ({})".format(D, N))



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Core functions for model descriptors

In spyks, models are specified by a map with following required keys: name
(string), state (mapping), equations (mapping), and parameters (mapping). Optional
keys include forcing (mapping) and reset (with predicate and state).

The convention in this package is that "mapping" is a list of (name, value)
tuples. Values must be specified with physical units (e.g., '5 mV') to permit
dimensional analysis; if units are not specified the quantity is assumed to be
dimensionless. An array, on the other hand, does not have names or quantities.

"""
# python 3 compatibility
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import os
import sympy as sp
import pint as pq

# need a global unit registry in order for quantities to be compatible
ureg = pq.UnitRegistry()


def n_params(model):
    """ Number of parameters in model """
    return len(model["parameters"])


def n_state(model):
    """ Number of state variables in model """
    return len(model["equations"])


def n_forcing(model):
    """ Number of forcing terms in model """
    return len(model["forcing"])


def param_names(model):
    """ Names of parameters in model """
    return tuple(n for n, v in model["parameters"])


def state_names(model):
    """ Names of state variables in model """
    return tuple(n for n, v in model["state"])


def forcing_names(model):
    """ Names of forcing terms in model """
    return tuple(n for n, v in model["forcing"])


def symbols(model):
    lh = set()
    rh = set()
    for n, eq in model["equations"]:
        lh.add(sp.Symbol(n))
        rh.update(eq.free_symbols)
    try:
        reset = model["reset"]
        rh.update(reset["predicate"].free_symbols)
        for n, eq in reset["state"]:
            rh.update(eq.free_symbols)
        for n, eq in reset.get("clip", []):
            rh.update(eq.free_symbols)
    except KeyError:
        pass
    return lh, rh


def to_array(mapping):
    """ Converts named vector of quantities into a dimensionless array. """
    from numpy import fromiter
    return fromiter((v.magnitude for k, v in mapping), dtype='d', count=len(mapping))


def to_scaled_array(mapping):
    """ Converts named vector of quantities into a dimensionless array. """
    from numpy import fromiter
    return fromiter((v.to_base_units().magnitude for k, v in mapping), dtype='d', count=len(mapping))


def to_mapping(array, mapping):
    """ Binds values in array to names and units in mapping. Size of array must match mapping. """
    assert len(array) == len(mapping)
    return [(k, x * getattr(v, 'units', v)) for  (k, v), x in zip(mapping, array) ]


def parse_quantity(x):
    if isinstance(x, (float, int)):
        return x * ureg.dimensionless
    else:
        return ureg.parse_expression(x)


class kinetic_ode(object):
    """Represents an ODE for a 1st order kinetic process

    Initialize with alpha & beta or inf and tau expressions
    """
    _alpha = None
    _beta = None
    _inf = None
    _tau = None

    def __init__(self, n, **eqns):
        self._var = sp.Symbol(n)
        if "alpha" in eqns:
            self._alpha = sp.sympify(eqns["alpha"])
        if "beta" in eqns:
            self._beta = sp.sympify(eqns["beta"])
        if "inf" in eqns:
            self._inf = sp.sympify(eqns["inf"])
        if "tau" in eqns:
            self._tau = sp.sympify(eqns["tau"])
        if not ((self._alpha is not None and self._beta is not None) or
                (self._inf is not None and self._tau is not None)):
            raise ValueError("must specify alpha/beta or inf/tau")

    @property
    def expr(self):
        if (self._alpha is not None and self._beta is not None):
            return self._alpha * (1 - self._var) - self._beta * self._var
        else:
            return (self._inf - self._var) / self._tau

    @property
    def inf(self):
        return self._inf or self._alpha / (self.alpha + self.beta)

    @property
    def tau(self):
        return self._tau or 1 / (self.alpha + self.beta)

    @property
    def alpha(self):
        return self._alpha or (self.inf / self.tau)

    @property
    def beta(self):
        return self._beta or ((1 - self.inf) / self.tau)



def parse_equation(n, s):
    """Parse equations of motion to sympy expression.

    This function tries to be smart about different formulations of state
    variables ODEs. It can understand the alpha/beta specification and the
    inf/tau specification. Otherwise it will throw an error. Errors are also
    thrown on badly formed equations.

    """
    import ruamel.yaml as yaml
    try:
        if isinstance(s, (dict, yaml.comments.CommentedMap)):
            expr = kinetic_ode(n, **s).expr
        else:
            expr = sp.sympify(s)
        return (n, expr)
    except (TypeError, sp.SympifyError):
        raise ValueError("unable to parse equation spec for {}".format(n))


def parse(model):
    """Parse equations, variables, and parameters of model into sympy objects"""
    from itertools import starmap
    from copy import copy
    model['eqns_unparsed'] = copy(model['equations'])
    model['equations'] = list(starmap(parse_equation, model['equations'].items()))
    try:
        reset = model["reset"]
        reset["predicate"] = sp.sympify(reset["predicate"])
        reset["state"] = [(sp.Symbol(n), sp.sympify(s)) for n,s in reset["state"].items()]
        reset["clip"] = [(sp.Symbol(n), sp.sympify(s)) for n,s in reset["clip"].items()]
    except KeyError:
        pass
    # parse quantities and units
    for k in ('parameters', 'forcing', 'state'):
        if k in model:
            model[k] = [(param, parse_quantity(value)) for param, value in model[k].items()]
    return model


def _mapping_updater(new):
    """Returns an f that updates mapping elements with values in new, converting units as needed"""
    def f(kv):
        n, v = kv
        if n in new:
            return (n, parse_quantity(new[n]).to(v.units))
        else:
            return (n, v)
    return f


def load_model(doc, load_base=True):
    """Loads a model descriptor from a file, string, or stream

    If the model descriptor has a "base" field, by default this function will
    use the supplied model to update the base model's default parameter and
    state values. If load_base is not True, then the function will raise a
    ValueError.

    """
    import ruamel.yaml as yaml
    if os.path.exists(doc):
        fname = doc
        fp = open(doc, "r")
        model = yaml.load(fp, yaml.RoundTripLoader)
        fp.close()
    else:
        fname = ""
        model = yaml.load(doc, yaml.RoundTripLoader)
    if "base" in model:
        if not load_base:
            raise ValueError("model extends %s but load_base is False" % model["base"])
        basefile = os.path.join(os.path.dirname(fname), model["base"] + ".yml")
        base = load_model(basefile)              # will get parsed
        if 'parameters' in model:
            base['parameters'] = list(
                map(_mapping_updater(dict(model['parameters'])), base['parameters']))
        if 'state' in model:
            base['state'] = list(
                map(_mapping_updater(dict(model['state'])), base['state']))
        return base
    else:
        return parse(model)


def update_model(model, **kwargs):
    """Updates model with arrays of values.

    Keyword arguments (e.g., parameters, forcing, state) must match existing
    keys, and array size must match size of model vector.

    """
    for key, array in kwargs.items():
        model[key] = to_mapping(array, model[key])


def get_param_value(model, name):
    """ Return parameter value from model """
    return dict(model["parameters"]).get(name)


def set_param_value(model, name, value):
    """ Update parameter value in the model """
    # this is somewhat clunky because the parameters are in a list, not a dict
    for i, (param, old) in enumerate(model["parameters"]):
        if param == name:
            model["parameters"][i] = (param, value)
            break


def load_module(model, path=None):
    """Loads the extension module associated with model.

    If `path` of extension module file is supplied, temporarily adds it to
    sys.path. Raises an error if the name or version don't match.

    """
    import sys
    import importlib
    if path is not None:
        sys.path.append(path)
    try:
        mdl = importlib.import_module(model["name"])
        if mdl.name != model["name"]:
            raise ImportError("extension module name ({}) doesn't match model name ({})".format(mdl.name,
                                                                                                model["name"]))
        if mdl.__version__ != model["version"]:
            raise ImportError("extension module version ({}) doesn't match descriptor ({})".format(mdl.__version__,
                                                                                                   model["version"]))
    finally:
        if path is not None:
            sys.path.remove(path)
    return mdl


def currents(model):
    """Extract intrinsic current terms from the model

    This function assumes that the model is biophysical: the model's first
    equation is for voltage, specified by current conservation.

    """
    dV = model["equations"][0][1]
    I_net = sp.fraction(dV)[0]
    forcing_vars = forcing_names(model)
    return (term for term in I_net.args if str(term) not in forcing_vars)


def conductances(model):
    """Extract intrinsic conductance terms from the model

    This function works by dividing out any arguments from currents that depend on the voltage
    """
    V = sp.Symbol("V")
    return (sp.prod(t for t in term.args if V not in t.free_symbols) for term in currents(model))


def conductance_names(model):
    """Extract names of conductances in the model"""
    return (str(expr.args[0]) for expr in currents(model))



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Functions for building extension modules """

# python 3 compatibility
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os
import sys
import sysconfig
import pybind11
import distutils
import logging
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

log = logging.getLogger('spyks')   # root logger


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        return pybind11.get_include(self.user)


def get_boost_include():
    if sys.platform == "win32" :
        return "C:/Boost/include/boost-1_32"
    elif sys.platform == 'darwin':
        return "/opt/local/include"
    else:
        return "/usr/include/boost"


def get_include(*args, **kwargs):
    try:
        from pip import locations
        return os.path.dirname(
            locations.distutils_scheme('spyks', *args, **kwargs)['headers'])
    except ImportError:
        return 'include'


def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    from distutils.errors import CompileError
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp', delete=True) as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class ImportCppExt(Extension):
    """ A custom Extension class for specifying installation destination """
    def __init__(self, libdest, *args, **kwargs):
        self.libdest = libdest
        Extension.__init__(self, *args, **kwargs)


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
            if has_flag(self.compiler, '-ffast-math'):
                opts.append('-ffast-math')
            if has_flag(self.compiler, '-flto'):
                opts.append('-flto')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


class CustomDestBuildExt(BuildExt):
    """A custom build extension for specifying target path"""
    def copy_extensions_to_source(self):
        for ext in self.extensions:
            fullname = self.get_ext_fullname(ext.name)
            filename = self.get_ext_filename(fullname)
            src_filename = os.path.join(self.build_lib, filename)
            dest_filename = os.path.join(ext.libdest, os.path.basename(filename))

            distutils.file_util.copy_file(
                src_filename, dest_filename,
                verbose = self.verbose, dry_run = self.dry_run
            )


def should_rebuild(modelfile, cppfile):
    t_model = os.path.getmtime(modelfile)
    t_cpp   = os.path.getmtime(cppfile)
    return t_cpp < t_model


def write_cppfile(model, fname, simplify=True):
    """ Renders model to C++ code and writes to fname"""
    from .codegen import render, simplify_equations
    if simplify:
        log.info("%s: simplifying equations", model["name"])
        model = simplify_equations(model)
    code = render(model)
    with open(fname, "wt") as fp:
        fp.write(code)
    log.info("%s: wrote %s", model["name"], fname)


def build_module(cppfile, name, path, **kwargs):
    import shutil, tempfile
    ext = ImportCppExt(
        path,
        name,
        language = "c++",
        sources = [cppfile],
        include_dirs = [get_pybind_include(), get_pybind_include(True),
                        get_boost_include(), get_include()],
    )

    build_path = tempfile.mkdtemp()
    args = ['build_ext', '--inplace', '-q']
    args.append('--build-temp=' + build_path)
    args.append('--build-lib=' + build_path)
    log.info("%s: compiling extension module", name)
    setup(name = name,
          version = kwargs.get("version", "0.0.1-SNAPSHOT"),
          ext_modules = [ext],
          script_args = args,
          cmdclass = { "build_ext": CustomDestBuildExt })
    shutil.rmtree(build_path)
    log.info("%s: complete - extension module in %s/", name, path)


def jit(model):
    """Compile a model file into an extension module and load it.

    Compilation can be quite slow for complex models, and this function doesn't
    do any kind of caching, so it's more for illustration purposes.

    """
    import spyks.validate as spkv
    spkv.check_symbols(model)
    spkv.check_equations(model)


def compile_script(argv=None):
    from spyks import __version__
    from spyks.core import load_model
    import spyks.validate as spkv
    import argparse

    p = argparse.ArgumentParser(description="compile a spyks model file into a python extension module")
    p.add_argument("model", help="the model descriptor file to compile")
    p.add_argument("target", help="the path to put the module (default same as model file)", nargs='?')
    p.add_argument("--version", action="version", version="%(prog)s {}".format(__version__))
    p.add_argument("--skip-compile", help="skip compilation step", action="store_true")
    p.add_argument("--skip-codegen", help="skip code generation step", action="store_true")
    args = p.parse_args(argv)

    ch = logging.StreamHandler()
    formatter = logging.Formatter("[%(name)s] %(message)s")
    loglevel = logging.INFO
    log.setLevel(loglevel)
    ch.setLevel(loglevel)  # change
    ch.setFormatter(formatter)
    log.addHandler(ch)

    try:
        model = load_model(args.model, load_base=False)
    except ValueError as e:
        log.info("error parsing model: %s", e)
        #log.info("%s extends a base model - compile the base model instead", args.model)
        return 0
    log.info("%s: validating model", model["name"])
    spkv.check_symbols(model)
    spkv.check_equations(model)
    path = args.target or os.path.dirname(args.model)
    cppfile = os.path.join(path, model["name"] + ".cpp")
    if not args.skip_codegen:
        write_cppfile(model, cppfile)

    if not args.skip_compile:
        build_module(cppfile, model["name"], path, version=model["version"])



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Generates equations.txt and parameters.txt for makecode (not implemented)"""
# python 3 compatibility
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import logging
from spyks.core import n_params, n_state, n_forcing
from spyks.codegen import simplify_equations

log = logging.getLogger('spyks')   # root logger


def discretize(model):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- mode: python -*-
from spyks import __version__
from setuptools import setup
import sys
if sys.version_info[:2] < (2, 7) or (3, 0) <= sys.version_info[:2] < (3, 5):
    raise RuntimeError("Python version 2.7 or >= 3.5 required.")


cls_txt = """
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
License :: OSI Approved :: GNU General Public License (GPL)
Programming Language :: Python
Topic :: Scientific/Engineering
Operating System :: Unix
Operating System :: POSIX :: Linux
Operating System :: MacOS :: MacOS X
Natural Language :: English
"""

long_desc = """
spyks is a tool for building and simulating simple, dynamical neuron models. It
places a strong emphasis on efficient modification of parameters, because it's
primarily intended for use in data assimilation applications where the goal is
to estimate parameters and unmeasured states from recorded data. It is very much
a work in progress.
"""

setup(
    name="spyks",
    version=__version__,
    description="Generate fast c++ code to integrate neuron models",
    long_description=long_desc,
    classifiers=[x for x in cls_txt.split("\n") if x],

    packages=['spyks'],
    package_data={'spyks': ["templates/*.cpp"]},
    entry_points={'console_scripts': ['spykscc = spyks.build:compile_script']},
    install_requires=[
        "pybind11>=2.2.0",
        "Pint>=0.7",
        "ruamel.yaml>=0.13",
        "sympy>=1.0",
        "numpy>=1.11"
    ],

    author="Tyler Robbins",
    maintainer='C Daniel Meliza',
    url="https://github.com/melizalab/spyks",
    test_suite='tests'
)



# -*- coding: utf-8 -*-
# -*- mode: python -*-

import unittest
import numpy as np

import spyks.core as s
import spyks.validate as sv

passive_params = [
    {'params': [57, 5.5, -70., 0., -65.],
     'forcing': [[0., 0.]],
     'state': [-70.]},
    {'params': (57, 5.5, -70., 0., -65.),
     'forcing': [[1.0, 0.]],
     'state': [-75.]},
    {'params': np.asarray([57, 5.5, -70., 0., -65.]),
     'forcing': [[0., 2.0]],
     'state': [-50.]},
]

class TestPassiveModel(unittest.TestCase):
    def test_passive_ei_deriv(self):
        pymodel = s.load_model("models/passive_ei.yml")
        passive_ei = s.load_module(pymodel, "models")
        def compare(params, forcing, state):
            model = passive_ei.model(params, forcing, 0.05)
            dXdt = model(state, 0)
            s.update_model(pymodel, state=state, forcing=forcing[0], parameters=params)
            pydXdt = s.to_array(sv.deriv(pymodel))
            self.assertTrue(np.allclose(dXdt, pydXdt))
        for tvals in passive_params:
            yield compare, tvals['params'], tvals['forcing'], tvals['state']

    def test_passive_ei_integration(self):
        pymodel = s.load_model("models/passive_ei.yml")
        passive_ei = s.load_module(pymodel, "models")
        g_ex = 1.0
        g_inh = 0.5
        N = 2000
        dt = 0.05
        params = s.to_array(pymodel['parameters'])
        forcing = np.column_stack([np.ones(N) * g_ex, np.ones(N) * g_inh])
        x0 = s.to_array(pymodel['state'])
        model = passive_ei.model(params, forcing, dt)
        X = passive_ei.integrate(params, x0, forcing, dt, dt)
        X2 = passive_ei.integrate(model, x0, N * dt, dt)
        self.assertTrue(np.allclose(X, X2))
        # steady-state V should be a weighted average of the reversal potentials
        g_l = s.get_param_value(pymodel, 'g_l')
        Evals = [s.get_param_value(pymodel, n) for n in ('E_l', 'E_ex', 'E_inh')]
        gvals = [g_l, g_ex * g_l.units, g_inh * g_l.units]
        V_steady = sum(g * e for e, g in zip(Evals, gvals)) / sum(gvals)
        events = (X[:,0] > 29.9).nonzero()[0]
        self.assertAlmostEqual(V_steady.magnitude, X[-1, 0], places=2)


adex_params = [
    {'params': [250, 30, -70.6, 2.0, -55, 144, 4, -70.6, 80.5, 30],
     'forcing': [0.],
     'state': [-70., 0]},
    {'params': (200, 32, -65, 2.0, -56, 120, 5, -70.6, 80.5, 30),
     'forcing': [0.05],
     'state': [-50., 0]},
    {'params': np.asarray([200, 32, -65, 2.0, -56, 120, 5, -70.6, 80.5, 30]),
     'forcing': [0.05],
     'state': [50., 10]},
]

class TestAdExModel(unittest.TestCase):
    def test_adex_dxdt(self):
        pymodel = s.load_model("models/adex.yml")
        adex = s.load_module(pymodel, "models")
        def compare_adex(params, forcing, state):
            # inj = models.timeseries(forcing, 0.05)
            model = adex.model(params, forcing, 0.05)
            dXdt = model(state, 0)
            s.update_model(pymodel, state=state, forcing=forcing, parameters=params)
            pydXdt = s.to_array(sv.deriv(pymodel))
            self.assertTrue(np.allclose(dXdt, pydXdt))
        for tvals in adex_params:
            yield compare_adex, tvals['params'], tvals['forcing'], tvals['state']

    def test_adex_reset(self):
        pymodel = s.load_model("models/adex.yml")
        adex = s.load_module(pymodel, "models")
        def compare_reset(params, forcing, state):
            model = adex.model(params, forcing, 0.05)
            reset, new_state = model.reset(state)
            s.update_model(pymodel, state=state, forcing=forcing, parameters=params)
            pypost_state = sv.reset(pymodel)
            if reset:
                pypost_state = s.to_array(pypost_state)
                self.assertTrue(np.allclose(new_state, pypost_state))
            else:
                self.assertEqual(reset, pypost_state)
                self.assertTrue(np.allclose(state, new_state))
        for tvals in adex_params:
            yield compare_reset, tvals['params'], tvals['forcing'], tvals['state']

    def test_adex_integration(self):
        pymodel = s.load_model("models/adex.yml")
        adex = s.load_module(pymodel, "models")
        I = 500
        N = 1000
        dt = 0.05
        params = adex_params[0]['params']
        x0 = adex_params[0]['state']
        data = np.ones(N) * I
        model = adex.model(params, data, dt)
        X = adex.integrate(params, x0, data, dt, dt)
        X2 = adex.integrate(model, x0, data.size * dt, dt)
        self.assertTrue(np.allclose(X, X2))
        # with these parameters, there should be exactly one spike at 555
        events = (X[:,0] > 29.9).nonzero()[0]
        self.assertEqual(events.size, 1)
        self.assertEqual(events[0], 555)
        return X


nakl_params = [
    {'params': [1., 120., 50., 20., -77., 0.3, -54.4,
                -40., 15., 0.1, 0.4, -40., 15., -60., -15.,
                1., 7., -60., -15., -55., 30., 1., 5., -55., -30.],
     'forcing': [0.],
     'state': [-70., 0., 0., 0.]},
    {'params': [1., 120., 50., 20., -77., 0.3, -54.4,
                -40., 15., 0.1, 0.4, -40., 15., -60., -15.,
                1., 7., -60., -15., -55., 30., 1., 5., -55., -30.],
     'forcing': [100.],
     'state': [-70., 0., 0.2, 0.1]},
    {'params': [1.02, 70.7, 55., .38, -85., 0.054, -65,
                -40., 15., 0.1, 0.4, -40., 15., -60., -15.,
                1., 7., -60., -15., -55., 30., 1., 5., -55., -30.],
     'forcing': [0.],
     'state': [-65., 0.1, 0.2, 0.1]},
]

class TestNaklModel(unittest.TestCase):
    def test_nakl_dxdt(self):
        pymodel = s.load_model("models/nakl.yml")
        nakl = s.load_module(pymodel, "models")
        def compare_nakl(params, forcing, state):
            model = nakl.model(params, forcing, 0.05)
            dXdt = model(state, 0)
            s.update_model(pymodel, state=state, forcing=forcing, parameters=params)
            pydXdt = s.to_array(sv.deriv(pymodel))
            self.assertTrue(np.allclose(dXdt, pydXdt))
        for tvals in nakl_params:
            yield compare_nakl, tvals['params'], tvals['forcing'], tvals['state']


    def test_nakl_integration(self, I=50):
        pymodel = s.load_model("models/nakl.yml")
        nakl = s.load_module(pymodel, "models")
        N = 10000
        dt = 0.05
        params = nakl_params[0]['params']
        x0 = nakl_params[0]['state']
        data = np.ones(N) * I
        model = nakl.model(params, data, dt)
        X = nakl.integrate(params, x0, data, dt, dt)
        X2 = nakl.integrate(model, x0, data.size * dt, dt)
        self.assertTrue(np.allclose(X, X2))
        return X


class TestBiocmModel(unittest.TestCase):
    def test_biocm(self):
        pymodel = s.load_model("models/biocm.yml")
        biocm = s.load_module(pymodel, "models")
        params = s.to_array(pymodel['parameters'])
        forcing = s.to_array(pymodel['forcing'])
        state = s.to_array(pymodel['state'])
        model = biocm.model(params, forcing, 0.05)
        dXdt = model(state.tolist(), 0)
        pydXdt = s.to_array(sv.deriv(pymodel))
        self.assertTrue(np.allclose(dXdt, pydXdt))


    def test_biocm_integration(self, I=20):
        N = 5000
        dt = 0.05
        pymodel = s.load_model("models/biocm.yml")
        biocm = s.load_module(pymodel, "models")
        params = s.to_array(pymodel['parameters'])
        x0 = s.to_array(pymodel['state'])
        data = np.ones(N) * I
        model = biocm.model(params, data, dt)
        X = biocm.integrate(params, x0, data, dt, dt)
        X2 = biocm.integrate(model, x0, data.size * dt, dt)
        self.assertTrue(np.allclose(X, X2))
        return X
